from tesserocr import PyTessBaseAPI  # Requires Tesseract 0.3.4+


def read_word(image, whitelist=None, chars=None, spaces=False):
    """ OCR a single word from an image. Useful for captchas.
        Image should be pre-processed to remove noise etc. """
    api = PyTessBaseAPI()
    api.SetPageSegMode(8)
    if whitelist is not None:
        api.SetVariable("tessedit_char_whitelist", whitelist)
    api.SetImage(image)
    api.Recognize()
    guess = api.GetUTF8Text()

    if not spaces:
        guess = ''.join([c for c in guess if c != " "])
        guess = guess.strip()

    if chars is not None and len(guess) != chars:
        return guess, None

    return guess, api.MeanTextConf()



import logging

from memorious.model.common import Base, unpack_int, pack_now
from memorious.util import make_key

log = logging.getLogger(__name__)


class CrawlerRun(Base):
    """The current state of a running crawler instance"""

    @classmethod
    def record_operation_start(cls, crawler, run_id):
        if not cls.conn.sismember(make_key(crawler, "runs"), run_id):
            cls.conn.sadd(make_key(crawler, "runs"), run_id)
            cls.conn.lpush(make_key(crawler, "runs_list"), run_id)
            cls.conn.set(make_key("run", run_id, "start"), pack_now())
        cls.conn.incr(make_key("run", run_id))
        cls.conn.incr(make_key("run", run_id, "total_ops"))

    @classmethod
    def record_operation_end(cls, crawler, run_id):
        cls.conn.decr(make_key("run", run_id))
        if unpack_int(cls.conn.get(make_key("run", run_id))) == 0:
            cls.conn.set(make_key("run", run_id, "end"), pack_now())

    @classmethod
    def flush(cls, crawler):
        for run_id in cls.conn.smembers(make_key(crawler, "runs")):
            cls.conn.delete(make_key("run", run_id, "start"))
            cls.conn.delete(make_key("run", run_id, "end"))
            cls.conn.delete(make_key("run", run_id, "total_ops"))
            cls.conn.delete(make_key("run", run_id))
        cls.conn.delete(make_key(crawler, "runs"))
        cls.conn.delete(make_key(crawler, "runs_list"))



from datetime import datetime

from memorious.core import connect_redis


def unpack_int(value):
    try:
        return int(value)
    except Exception:
        return 0


def pack_datetime(value):
        if value is not None:
            return str(value)


def pack_now():
    return pack_datetime(datetime.utcnow())


def unpack_datetime(value):
    if value is not None:
        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")


class Base(object):
    conn = connect_redis()



import json
import logging

from memorious.model.common import Base, pack_now, unpack_datetime
from memorious.util import make_key

log = logging.getLogger(__name__)


class Event(Base):
    """Document errors and warnings caused during tasks."""
    LEVEL_WARNING = 'warning'
    LEVEL_ERROR = 'error'
    LEVELS = [LEVEL_WARNING, LEVEL_ERROR]

    @classmethod
    def save(cls, crawler, stage, level, run_id, error=None, message=None):
        """Create an event, possibly based on an exception."""
        event = {
            'stage': stage.name,
            'level': level,
            'timestamp': pack_now(),
            'error': error,
            'message': message
        }
        data = json.dumps(event)
        cls.conn.lpush(make_key(crawler, "events"), data)
        cls.conn.lpush(make_key(crawler, "events", level), data)
        cls.conn.lpush(make_key(crawler, stage, "events"), data)
        cls.conn.lpush(make_key(crawler, stage.name, "events", level), data)
        cls.conn.lpush(make_key(crawler, run_id, "events"), data)
        cls.conn.lpush(make_key(crawler, run_id, "events", level), data)
        return event

    @classmethod
    def delete(cls, crawler):
        cls.conn.delete(make_key(crawler, "events"))
        for level in cls.LEVELS:
            cls.conn.delete(make_key(crawler, "events", level))
            for stage in crawler.stages:
                cls.conn.delete(make_key(crawler, stage, "events", level))

    @classmethod
    def get_counts(cls, crawler):
        counts = {}
        for level in cls.LEVELS:
            key = make_key(crawler, "events", level)
            counts[level] = cls.conn.llen(key) or 0
        return counts

    @classmethod
    def get_stage_counts(cls, crawler, stage):
        counts = {}
        for level in cls.LEVELS:
            key = make_key(crawler, stage, "events", level)
            counts[level] = cls.conn.llen(key) or 0
        return counts

    @classmethod
    def get_run_counts(cls, crawler, run_id):
        counts = {}
        for level in cls.LEVELS:
            key = make_key(crawler, run_id, "events", level)
            counts[level] = cls.conn.llen(key) or 0
        return counts

    @classmethod
    def event_list(cls, key, start, end):
        results = []
        events = cls.conn.lrange(key, start, end)
        if events is None:
            return results
        for event in events:
            result = json.loads(event)
            result["timestamp"] = unpack_datetime(result['timestamp'])
            results.append(result)
        return results

    @classmethod
    def get_crawler_events(cls, crawler, start, end, level=None):
        key = make_key(crawler, "events", level)
        return cls.event_list(key, start, end)

    @classmethod
    def get_stage_events(cls, crawler, stage_name, start, end, level=None):
        """events from a particular stage"""
        key = make_key(crawler, stage_name, "events", level)
        return cls.event_list(key, start, end)

    @classmethod
    def get_run_events(cls, crawler, run_id, start, end, level=None):
        """Events from a particular run"""
        key = make_key(crawler, run_id, "events", level)
        return cls.event_list(key, start, end)



import os
import logging
from pkg_resources import iter_entry_points

import redis
import fakeredis
import dataset
import storagelayer
from celery import Celery
from celery.schedules import crontab
from werkzeug.local import LocalProxy
from raven import Client
from raven.contrib.celery import register_signal, register_logger_signal

from memorious import settings
from memorious.logic.queue import CrawlerExecutionQueue


log = logging.getLogger(__name__)

celery = Celery('memorious')
celery.conf.update(
    imports=('memorious.tasks'),
    broker_url=settings.BROKER_URI,
    broker_transport_options={'fanout_prefix': True},
    task_always_eager=settings.EAGER,
    task_eager_propagates=True,
    task_ignore_result=True,
    task_default_queue=settings.APP_NAME,
    task_default_routing_key='%s.process' % settings.APP_NAME,
    result_persistent=False,
    worker_max_tasks_per_child=1000,
    # worker_prefetch_multiplier=10,
    # worker_hijack_root_logger=False,
    beat_schedule={
        'scheduled-crawlers': {
            'task': 'memorious.tasks.process_schedule',
            'schedule': crontab(minute='*/30')
        },
        'cleanup-crawlers': {
            'task': 'memorious.tasks.run_cleanup',
            'schedule': crontab(hour='*')
        },
    },
)

redis_pool = redis.ConnectionPool(
    host=settings.REDIS_HOST,
    port=settings.REDIS_PORT,
    decode_responses=True
)

# set up a task queue using a Queue if celery is set to eager mode.
local_queue = CrawlerExecutionQueue()

# set up raven for error reporting
if settings.SENTRY_DSN:
    client = Client(settings.SENTRY_DSN)
    register_logger_signal(client)
    register_signal(client, ignore_expected=True)


# File storage layer for blobs on local file system or S3
storage = storagelayer.init(settings.ARCHIVE_TYPE,
                            path=settings.ARCHIVE_PATH,
                            aws_key_id=settings.ARCHIVE_AWS_KEY_ID,
                            aws_secret=settings.ARCHIVE_AWS_SECRET,
                            aws_region=settings.ARCHIVE_AWS_REGION,
                            bucket=settings.ARCHIVE_BUCKET)


def load_manager():
    if not hasattr(settings, '_manager'):
        from memorious.logic.manager import CrawlerManager
        settings._manager = CrawlerManager()
        if settings.CONFIG_PATH:
            settings._manager.load_path(settings.CONFIG_PATH)
    return settings._manager


def load_datastore():
    if not hasattr(settings, '_datastore'):
        if not settings.DATASTORE_URI:
            raise RuntimeError("No $MEMORIOUS_DATASTORE_URI.")
        settings._datastore = dataset.connect(settings.DATASTORE_URI)
        # Use bigint to store integers by default
        settings._datastore.types.integer = settings._datastore.types.bigint
    return settings._datastore


manager = LocalProxy(load_manager)
datastore = LocalProxy(load_datastore)


def connect_redis():
    if not settings.REDIS_HOST:
        return fakeredis.FakeRedis(decode_responses=True)
    return redis.Redis(connection_pool=redis_pool, decode_responses=True)


def load_extensions():
    for ep in iter_entry_points('memorious.plugins'):
        func = ep.load()
        func()


def init_memorious():
    load_extensions()



import json
import logging

from memorious.model.common import Base
from memorious.util import make_key

log = logging.getLogger(__name__)


class Tag(Base):
    """A simple key/value table used to store interim results."""

    @classmethod
    def save(cls, crawler, key, value):
        data = json.dumps(value)
        cls.conn.set(make_key(crawler, "tag", key), data, ex=crawler.expire)

    @classmethod
    def find(cls, crawler, key):
        value = cls.conn.get(make_key(crawler, "tag", key))
        if value is not None:
            return json.loads(value)

    @classmethod
    def exists(cls, crawler, key):
        return cls.conn.exists(make_key(crawler, "tag", key))

    @classmethod
    def delete(cls, crawler):
        for key in cls.conn.scan_iter(make_key(crawler, "tag", "*")):
            cls.conn.delete(key)



import os
import pkg_resources
import six
from banal import as_bool


def env(name, default=None, required=False):
    name = 'MEMORIOUS_%s' % name.upper().strip()
    if required and name not in os.environ:
        raise RuntimeError("Missing configuration: $%s" % name)
    return os.environ.get(name, default)


def env_bool(name, default=False):
    """Extract a boolean value from the environment consistently."""
    return as_bool(env(name), default=default)


###############################################################################
# Core configuration

VERSION = pkg_resources.get_distribution('memorious').version
APP_NAME = env('APP_NAME', 'memorious')

# Enable debug logging etc.
DEBUG = env_bool('DEBUG', default=False)

# Base operating path
BASE_PATH = env('BASE_PATH', os.path.join(six.moves.getcwd(), 'data'))

# Directory which contains crawler pipeline YAML specs
CONFIG_PATH = env('CONFIG_PATH')

# Try and run scrapers in a way that only acquires new data
INCREMENTAL = env_bool('INCREMENTAL', default=True)

# How many days until an incremental crawl expires
EXPIRE = int(env('EXPIRE', 60))

# HTTP request configuration
HTTP_CACHE = env_bool('HTTP_CACHE', default=True)

# HTTP user agent default
USER_AGENT = 'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.1)'
USER_AGENT = '%s aleph.memorious/%s' % (USER_AGENT, VERSION)
USER_AGENT = env('USER_AGENT', USER_AGENT)

# Datastore: operational data store (ODS) database connection
DATASTORE_FILE = os.path.join(BASE_PATH, 'datastore.sqlite3')
DATASTORE_URI = env('DATASTORE_URI', 'sqlite:///%s' % DATASTORE_FILE)


###############################################################################
# Queue processing

BROKER_URI = 'amqp://guest:guest@localhost:5672//'
BROKER_URI = env('BROKER_URI', BROKER_URI)

# Enable delayed processing via queue
EAGER = env_bool('EAGER', True)


###############################################################################
# Data storage

# Archive type (either 's3' or 'file', i.e. local file system):
ARCHIVE_TYPE = env('ARCHIVE_TYPE', 'file')
ARCHIVE_PATH = env('ARCHIVE_PATH', os.path.join(BASE_PATH, 'archive'))
ARCHIVE_AWS_KEY_ID = env('AWS_ACCESS_KEY_ID',
                         os.environ.get('AWS_ACCESS_KEY_ID'))
ARCHIVE_AWS_SECRET = env('AWS_SECRET_ACCESS_KEY',
                         os.environ.get('AWS_SECRET_ACCESS_KEY'))
ARCHIVE_AWS_REGION = env('ARCHIVE_REGION', 'eu-west-1')
ARCHIVE_BUCKET = env('ARCHIVE_BUCKET')


###############################################################################
# Aleph module

ALEPH_HOST = env('ALEPH_HOST')
ALEPH_API_KEY = env('ALEPH_API_KEY')


###############################################################################
# Error reporting

# Using sentry raven
SENTRY_DSN = env('SENTRY_DSN')


###############################################################################
# Redis

REDIS_HOST = env('REDIS_HOST', '')
REDIS_PORT = int(env('REDIS_PORT', 6379))



import logging
from threading import Thread
from six.moves.queue import Queue


log = logging.getLogger(__name__)


class CrawlerExecutionQueue(object):
    """Queue and execute operations in a separate thread when celery is
    running in eager mode.
    """
    def __init__(self):
        self.queue = Queue()

    def queue_operation(self, context, data):
        if not hasattr(self, 'worker'):
            self.worker = Thread(target=self.execute_crawler)
            self.worker.setDaemon(True)
            self.worker.start()
        self.queue.put((context, data))

    def execute_crawler(self):
        while True:
            context, data = self.queue.get()
            context.execute(data)
            self.queue.task_done()

    @property
    def is_empty(self):
        return not self.queue.unfinished_tasks



import re
import six
from six.moves.urllib.parse import urlparse
from celestial import normalize_mimetype

from memorious.logic.mime import GROUPS


class Rule(object):
    def __init__(self, value):
        self.value = value

    def configure(self):
        pass

    def apply(self, res):
        raise NotImplemented()

    def to_dict(self):
        return self.value

    @staticmethod
    def get_rule(spec):
        if not isinstance(spec, dict):
            raise Exception('Not a valid rule: %r' % spec)
        if len(spec) > 1:
            raise Exception('Ambiguous rules: %r' % spec)
        for rule_name, value in spec.items():
            rule_cls = RULES.get(rule_name)
            if rule_cls is None:
                raise Exception('Unknown rule: %s' % rule_name)
            rule = rule_cls(value)
            rule.configure()
            return rule
        raise Exception('Empty rule: %s' % spec)


class ListRule(Rule):
    """An abstract type of rules that contain a set of other rules."""

    def configure(self):
        if not isinstance(self.value, (list, set, tuple)):
            raise Exception("Not a list: %r", self.value)

    @property
    def children(self):
        for rule in self.value:
            yield self.get_rule(rule)


class OrRule(ListRule):
    """Any nested rule must apply."""

    def apply(self, res):
        for rule in self.children:
            if rule.apply(res):
                return True
        return False


class AndRule(ListRule):
    """All nested rules must apply."""

    def apply(self, res):
        for rule in self.children:
            if not rule.apply(res):
                return False
        return True


class NotRule(Rule):
    """Invert a nested rule."""

    def configure(self):
        self.rule = self.get_rule(self.value)

    def apply(self, res):
        return not self.rule.apply(res)


class MatchAllRule(Rule):
    """Just say yes."""

    def apply(self, res):
        return True


class MimeTypeRule(Rule):

    def configure(self):
        self.clean = normalize_mimetype(self.value)

    def apply(self, res):
        return self.res.content_type == self.clean


class MimeGroupRule(Rule):

    def apply(self, res):
        if res.content_type.startswith('%s/' % self.value):
            return True
        return res.content_type in GROUPS.get(self.value, [])


class DomainRule(Rule):
    """Match all pages from a particular domain."""

    def clean_domain(self, domain):
        if domain is None:
            return
        pr = urlparse(domain)
        domain = pr.hostname or pr.path
        domain = domain.strip('.').lower()
        return domain

    def configure(self):
        if not isinstance(self.value, six.string_types):
            raise Exception("Not a domain: %r", self.value)
        self.domain = self.clean_domain(self.value)
        self.sub_domain = '.%s' % self.domain

    def apply(self, res):
        hostname = self.clean_domain(res.url)
        if hostname is None or self.domain is None:
            return False
        if hostname == self.domain:
            return True
        if hostname.endswith(self.sub_domain):
            return True
        return False


class UrlPatternRule(Rule):

    def configure(self):
        if not isinstance(self.value, six.string_types):
            raise Exception("Not a regex: %r", self.value)
        self.pattern = re.compile(self.value, re.I | re.U)

    def apply(self, res):
        if self.pattern.match(res.url):
            return True
        return False


RULES = {}
RULES['or'] = OrRule
RULES['any'] = OrRule
RULES['and'] = AndRule
RULES['all'] = AndRule
RULES['not'] = NotRule
RULES['match_all'] = MatchAllRule
RULES['domain'] = DomainRule
RULES['mime_type'] = MimeTypeRule
RULES['mime_group'] = MimeGroupRule
RULES['pattern'] = UrlPatternRule



import datetime
from banal import ensure_list

from memorious.core import datastore


def _upsert(context, params, data):
    """Insert or update data and add/update appropriate timestamps"""
    table = params.get("table")
    table = datastore.get_table(table, primary_id=False)
    unique_keys = ensure_list(params.get("unique"))
    data["__last_seen"] = datetime.datetime.utcnow()
    if len(unique_keys):
        updated = table.update(data, unique_keys, return_count=True)
        if updated:
            return
    data["__first_seen"] = data["__last_seen"]
    table.insert(data)


def _recursive_upsert(context, params, data):
    """Insert or update nested dicts recursively into db tables"""
    children = params.get("children", {})
    nested_calls = []
    for child_params in children:
        key = child_params.get("key")
        child_data_list = ensure_list(data.pop(key))
        if isinstance(child_data_list, dict):
            child_data_list = [child_data_list]
        if not (isinstance(child_data_list, list) and
                all(isinstance(i, dict) for i in child_data_list)):
            context.log.warn(
                "Expecting a dict or a lost of dicts as children for key", key
            )
            continue
        if child_data_list:
            table_suffix = child_params.get("table_suffix", key)
            child_params["table"] = params.get("table") + "_" + table_suffix
            # copy some properties over from parent to child
            inherit = child_params.get("inherit", {})
            for child_data in child_data_list:
                for dest, src in inherit.items():
                    child_data[dest] = data.get(src)
                nested_calls.append((child_params, child_data))
    # Insert or update data
    _upsert(context, params, data)
    for child_params, child_data in nested_calls:
        _recursive_upsert(context, child_params, child_data)


def db(context, data):
    """Insert or update `data` as a row into specified db table"""
    table = context.params.get("table", context.crawler.name)
    params = context.params
    params["table"] = table
    _recursive_upsert(context, params, data)



import os

from memorious.logic.manager import Crawler
from memorious.logic.stage import CrawlerStage


class TestCrawler(object):
    def test_crawler(self, crawler_dir, manager):
        source_file = os.path.join(crawler_dir, "simple_web_scraper.yml")
        crawler = Crawler(manager, source_file)
        assert crawler.name == "occrp_web_site"
        assert crawler.schedule == "weekly"
        assert set(crawler.stages.keys()) == {"init", "fetch", "parse", "store"}
        assert all(
            isinstance(stage, CrawlerStage) for stage in crawler.stages.values()
        )



import os
import logging
from fnmatch import fnmatch

from memorious.logic.crawler import Crawler

log = logging.getLogger(__name__)


class CrawlerManager(object):
    """Crawl a directory of YAML files to load a set of crawler specs."""

    def __init__(self):
        self.crawlers = {}

    def load_path(self, path):
        if not os.path.isdir(path):
            log.warning('Crawler config path %s not found.', path)

        for root, _, file_names in os.walk(path):
            for file_name in file_names:
                if not (fnmatch(file_name, '*.yaml') or fnmatch(file_name, '*.yml')):  # noqa
                    continue
                source_file = os.path.join(root, file_name)
                crawler = Crawler(self, source_file)
                self.crawlers[crawler.name] = crawler

    def run_scheduled(self):
        log.info('Checking schedule: %s crawlers.' % len(self.crawlers))
        for crawler in self:
            if crawler.delta is None:
                log.info('[%s] has no schedule.', crawler.name)
                continue
            if not crawler.check_due():
                continue
            log.info('[%s] due.', crawler.name)
            crawler.run()

    def run_cleanup(self):
        log.info('Cleanup: %s crawlers.' % len(self.crawlers))
        for crawler in self:
            crawler.cleanup()

    def __getitem__(self, name):
        return self.crawlers.get(name)

    def __iter__(self):
        crawlers = list(self.crawlers.values())
        crawlers.sort(key=lambda c: c.name)
        return iter(crawlers)

    def __len__(self):
        return len(self.crawlers)

    def get(self, name):
        return self.crawlers.get(name)



from memorious.core import manager, celery as app
from memorious.core import init_memorious
from memorious.logic.context import handle  # noqa


init_memorious()


@app.task
def process_schedule():
    manager.run_scheduled()


@app.task
def run_cleanup():
    manager.run_cleanup()
